@@ -40,7 +40,7 @@
 
 #include "atmodem.h"
 
-#define STATIC_IP_NETMASK "255.255.255.248"
+#define STATIC_IP_NETMASK "255.255.255.255"
 
 static const char *none_prefix[] = { NULL };
 
@@ -65,7 +65,8 @@ struct gprs_context_data {
 	void *cb_data;                                  /* Callback data */
 };
 
-static void ppp_connect(const char *interface, const char *ip,
+static void ppp_connect(const char *interface, const char *local,
+			const char *remote,
 			const char *dns1, const char *dns2,
 			gpointer user_data)
 {
@@ -78,7 +79,7 @@ static void ppp_connect(const char *interface, const char *ip,
 	dns[2] = 0;
 
 	gcd->state = STATE_ACTIVE;
-	CALLBACK_WITH_SUCCESS(gcd->up_cb, interface, TRUE, ip,
+	CALLBACK_WITH_SUCCESS(gcd->up_cb, interface, TRUE, local,
 					STATIC_IP_NETMASK, NULL,
 					dns, gcd->cb_data);
 }
@@ -90,6 +91,8 @@ static void ppp_disconnect(GAtPPPDisconnectReason reason, gpointer user_data)
 
 	DBG("");
 
+	g_at_ppp_unref(gcd->ppp);
+	gcd->ppp = NULL;
 	g_at_chat_resume(gcd->chat);
 
 	switch (gcd->state) {
@@ -252,6 +255,13 @@ static void at_gprs_context_remove(struct ofono_gprs_context *gc)
 {
 	struct gprs_context_data *gcd = ofono_gprs_context_get_data(gc);
 
+	DBG("");
+
+	if (gcd->state != STATE_IDLE) {
+		g_at_ppp_unref(gcd->ppp);
+		g_at_chat_resume(gcd->chat);
+	}
+
 	ofono_gprs_context_set_data(gc, NULL);
 	g_free(gcd);
 }
