@@ -44,9 +44,14 @@
 struct sim_data {
 	GAtChat *chat;
 	unsigned int vendor;
+	guint epev_id;
+	guint epev_source;
 };
 
 static const char *crsm_prefix[] = { "+CRSM:", NULL };
+static const char *cpin_prefix[] = { "+CPIN:", NULL };
+static const char *clck_prefix[] = { "+CLCK:", NULL };
+static const char *none_prefix[] = { NULL };
 
 static void at_crsm_info_cb(gboolean ok, GAtResult *result, gpointer user_data)
 {
@@ -104,9 +109,19 @@ static void at_sim_read_info(struct ofono_sim *sim, int fileid,
 					void *data)
 {
 	struct sim_data *sd = ofono_sim_get_data(sim);
-	struct cb_data *cbd = cb_data_new(cb, data);
+	struct cb_data *cbd;
 	char buf[64];
 
+	if (sd->vendor == OFONO_VENDOR_OPTION_HSO) {
+		unsigned char access[3] = { 0x00, 0x00, 0x00 };
+
+		if (fileid == SIM_EFAD_FILEID) {
+			CALLBACK_WITH_SUCCESS(cb, 4, 0, 0, access, data);
+			return;
+		}
+	}
+
+	cbd = cb_data_new(cb, data);
 	if (!cbd)
 		goto error;
 
@@ -120,9 +135,6 @@ static void at_sim_read_info(struct ofono_sim *sim, int fileid,
 		return;
 
 error:
-	if (cbd)
-		g_free(cbd);
-
 	CALLBACK_WITH_FAILURE(cb, -1, -1, -1, NULL, data);
 }
 
@@ -404,22 +416,22 @@ static struct {
 	enum ofono_sim_password_type type;
 	const char *name;
 } const at_sim_name[] = {
-	{ OFONO_SIM_PASSWORD_NONE, "READY" },
-	{ OFONO_SIM_PASSWORD_SIM_PIN, "SIM PIN" },
-	{ OFONO_SIM_PASSWORD_SIM_PUK, "SIM PUK" },
-	{ OFONO_SIM_PASSWORD_PHSIM_PIN, "PH-SIM PIN" },
-	{ OFONO_SIM_PASSWORD_PHFSIM_PIN, "PH-FSIM PIN" },
-	{ OFONO_SIM_PASSWORD_PHFSIM_PUK, "PH-FSIM PUK" },
-	{ OFONO_SIM_PASSWORD_SIM_PIN2, "SIM PIN2" },
-	{ OFONO_SIM_PASSWORD_SIM_PUK2, "SIM PUK2" },
-	{ OFONO_SIM_PASSWORD_PHNET_PIN, "PH-NET PIN" },
-	{ OFONO_SIM_PASSWORD_PHNET_PUK, "PH-NET PUK" },
-	{ OFONO_SIM_PASSWORD_PHNETSUB_PIN, "PH-NETSUB PIN" },
-	{ OFONO_SIM_PASSWORD_PHNETSUB_PUK, "PH-NETSUB PUK" },
-	{ OFONO_SIM_PASSWORD_PHSP_PIN, "PH-SP PIN" },
-	{ OFONO_SIM_PASSWORD_PHSP_PUK, "PH-SP PUK" },
-	{ OFONO_SIM_PASSWORD_PHCORP_PIN, "PH-CORP PIN" },
-	{ OFONO_SIM_PASSWORD_PHCORP_PUK, "PH-CORP PUK" },
+	{ OFONO_SIM_PASSWORD_NONE,		"READY"		},
+	{ OFONO_SIM_PASSWORD_SIM_PIN,		"SIM PIN"	},
+	{ OFONO_SIM_PASSWORD_SIM_PUK,		"SIM PUK"	},
+	{ OFONO_SIM_PASSWORD_PHSIM_PIN,		"PH-SIM PIN"	},
+	{ OFONO_SIM_PASSWORD_PHFSIM_PIN,	"PH-FSIM PIN"	},
+	{ OFONO_SIM_PASSWORD_PHFSIM_PUK,	"PH-FSIM PUK"	},
+	{ OFONO_SIM_PASSWORD_SIM_PIN2,		"SIM PIN2"	},
+	{ OFONO_SIM_PASSWORD_SIM_PUK2,		"SIM PUK2"	},
+	{ OFONO_SIM_PASSWORD_PHNET_PIN,		"PH-NET PIN"	},
+	{ OFONO_SIM_PASSWORD_PHNET_PUK,		"PH-NET PUK"	},
+	{ OFONO_SIM_PASSWORD_PHNETSUB_PIN,	"PH-NETSUB PIN"	},
+	{ OFONO_SIM_PASSWORD_PHNETSUB_PUK,	"PH-NETSUB PUK"	},
+	{ OFONO_SIM_PASSWORD_PHSP_PIN,		"PH-SP PIN"	},
+	{ OFONO_SIM_PASSWORD_PHSP_PUK,		"PH-SP PUK"	},
+	{ OFONO_SIM_PASSWORD_PHCORP_PIN,	"PH-CORP PIN"	},
+	{ OFONO_SIM_PASSWORD_PHCORP_PUK,	"PH-CORP PUK"	},
 };
 
 static void at_cpin_cb(gboolean ok, GAtResult *result, gpointer user_data)
@@ -488,7 +500,7 @@ static void at_pin_query(struct ofono_sim *sim, ofono_sim_passwd_cb_t cb,
 
 	cbd->user = sim;
 
-	if (g_at_chat_send(sd->chat, "AT+CPIN?", NULL,
+	if (g_at_chat_send(sd->chat, "AT+CPIN?", cpin_prefix,
 				at_cpin_cb, cbd, g_free) > 0)
 		return;
 
@@ -499,6 +511,59 @@ error:
 	CALLBACK_WITH_FAILURE(cb, -1, data);
 }
 
+static gboolean at_epev_unregister(gpointer user_data)
+{
+	struct sim_data *sd = user_data;
+
+	sd->epev_source = 0;
+
+	g_at_chat_unregister(sd->chat, sd->epev_id);
+	sd->epev_id = 0;
+
+	return FALSE;
+}
+
+static void at_epev_notify(GAtResult *result, gpointer user_data)
+{
+	struct cb_data *cbd = user_data;
+	struct sim_data *sd = cbd->user;
+	ofono_sim_lock_unlock_cb_t cb = cbd->cb;
+	struct ofono_error error = { .type = OFONO_ERROR_TYPE_NO_ERROR };
+
+	if (sd->epev_source)
+		return;
+
+	cb(&error, cbd->data);
+
+	sd->epev_source = g_timeout_add(0, at_epev_unregister, sd);
+}
+
+static void at_pin_send_cb(gboolean ok, GAtResult *result,
+				gpointer user_data)
+{
+	struct cb_data *cbd = user_data;
+	struct sim_data *sd = cbd->user;
+	ofono_sim_lock_unlock_cb_t cb = cbd->cb;
+	struct ofono_error error;
+
+	decode_at_error(&error, g_at_result_final_response(result));
+
+	/*
+	 * On the MBM modem, AT+CPIN? keeps returning SIM PIN for a moment
+	 * after successful AT+CPIN="..", but sends *EPEV when that changes.
+	 */
+	if (ok && sd->vendor == OFONO_VENDOR_MBM) {
+		sd->epev_id = g_at_chat_register(sd->chat, "*EPEV",
+							at_epev_notify,
+							FALSE, cbd, g_free);
+		return;
+	}
+
+	cb(&error, cbd->data);
+
+	g_free(cbd);
+}
+
 static void at_lock_unlock_cb(gboolean ok, GAtResult *result,
 				gpointer user_data)
 {
@@ -522,10 +587,12 @@ static void at_pin_send(struct ofono_sim *sim, const char *passwd,
 	if (!cbd)
 		goto error;
 
+	cbd->user = sd;
+
 	snprintf(buf, sizeof(buf), "AT+CPIN=\"%s\"", passwd);
 
-	ret = g_at_chat_send(sd->chat, buf, NULL,
-				at_lock_unlock_cb, cbd, g_free);
+	ret = g_at_chat_send(sd->chat, buf, none_prefix,
+				at_pin_send_cb, cbd, NULL);
 
 	memset(buf, 0, sizeof(buf));
 
@@ -553,7 +620,7 @@ static void at_pin_send_puk(struct ofono_sim *sim, const char *puk,
 
 	snprintf(buf, sizeof(buf), "AT+CPIN=\"%s\",\"%s\"", puk, passwd);
 
-	ret = g_at_chat_send(sd->chat, buf, NULL,
+	ret = g_at_chat_send(sd->chat, buf, none_prefix,
 				at_lock_unlock_cb, cbd, g_free);
 
 	memset(buf, 0, sizeof(buf));
@@ -599,7 +666,7 @@ static void at_pin_enable(struct ofono_sim *sim,
 	snprintf(buf, sizeof(buf), "AT+CLCK=\"%s\",%i,\"%s\"",
 			at_clck_cpwd_fac[passwd_type], enable ? 1 : 0, passwd);
 
-	ret = g_at_chat_send(sd->chat, buf, NULL,
+	ret = g_at_chat_send(sd->chat, buf, none_prefix,
 				at_lock_unlock_cb, cbd, g_free);
 
 	memset(buf, 0, sizeof(buf));
@@ -635,7 +702,7 @@ static void at_change_passwd(struct ofono_sim *sim,
 	snprintf(buf, sizeof(buf), "AT+CPWD=\"%s\",\"%s\",\"%s\"",
 			at_clck_cpwd_fac[passwd_type], old, new);
 
-	ret = g_at_chat_send(sd->chat, buf, NULL,
+	ret = g_at_chat_send(sd->chat, buf, none_prefix,
 				at_lock_unlock_cb, cbd, g_free);
 
 	memset(buf, 0, sizeof(buf));
@@ -698,7 +765,7 @@ static void at_pin_query_enabled(struct ofono_sim *sim,
 	snprintf(buf, sizeof(buf), "AT+CLCK=\"%s\",2",
 			at_clck_cpwd_fac[passwd_type]);
 
-	if (g_at_chat_send(sd->chat, buf, NULL,
+	if (g_at_chat_send(sd->chat, buf, clck_prefix,
 				at_lock_status_cb, cbd, g_free) > 0)
 		return;
 
@@ -709,82 +776,6 @@ error:
 	CALLBACK_WITH_FAILURE(cb, -1, data);
 }
 
-static void at_csim_envelope_cb(gboolean ok, GAtResult *result,
-		gpointer user_data)
-{
-	struct cb_data *cbd = user_data;
-	GAtResultIter iter;
-	ofono_sim_read_cb_t cb = cbd->cb;
-	struct ofono_error error;
-	const guint8 *response;
-	gint rlen, len;
-
-	decode_at_error(&error, g_at_result_final_response(result));
-
-	if (!ok)
-		goto error;
-
-	g_at_result_iter_init(&iter, result);
-
-	if (!g_at_result_iter_next(&iter, "+CSIM:"))
-		goto error;
-
-	if (!g_at_result_iter_next_number(&iter, &rlen))
-		goto error;
-
-	if (!g_at_result_iter_next_hexstring(&iter, &response, &len))
-		goto error;
-
-	if (rlen != len * 2 || len < 2 ||
-			response[len - 2] != 0x90 || response[len - 1] != 0)
-		goto error;
-
-	DBG("csim_envelope_cb: %i", len);
-
-	cb(&error, response, len - 2, cbd->data);
-	return;
-
-error:
-	CALLBACK_WITH_FAILURE(cb, NULL, 0, cbd->data);
-}
-
-static void at_sim_envelope(struct ofono_sim *sim, int length,
-				const guint8 *command,
-				ofono_sim_read_cb_t cb, void *data)
-{
-	struct sim_data *sd = ofono_sim_get_data(sim);
-	struct cb_data *cbd = cb_data_new(cb, data);
-	char *buf = g_try_new(char, 64 + length * 2);
-	int len, ret;
-
-	if (!cbd || !buf)
-		goto error;
-
-	len = sprintf(buf, "AT+CSIM=%i,A0C20000%02hhX",
-			10 + length * 2, length);
-
-	for (; length; length--)
-		len += sprintf(buf + len, "%02hhX", *command++);
-
-	ret = g_at_chat_send(sd->chat, buf, crsm_prefix,
-				at_csim_envelope_cb, cbd, g_free);
-
-	g_free(buf);
-	buf = NULL;
-
-	if (ret > 0)
-		return;
-
-error:
-	if (buf)
-		g_free(buf);
-
-	if (cbd)
-		g_free(cbd);
-
-	CALLBACK_WITH_FAILURE(cb, NULL, 0, data);
-}
-
 static gboolean at_sim_register(gpointer user)
 {
 	struct ofono_sim *sim = user;
@@ -804,8 +795,16 @@ static int at_sim_probe(struct ofono_sim *sim, unsigned int vendor,
 	sd->chat = chat;
 	sd->vendor = vendor;
 
-	if (sd->vendor == OFONO_VENDOR_WAVECOM)
+	switch (sd->vendor) {
+	case OFONO_VENDOR_WAVECOM:
 		g_at_chat_add_terminator(chat, "+CPIN:", 6, TRUE);
+		break;
+	case OFONO_VENDOR_MBM:
+		g_at_chat_send(chat, "AT*EPEE=1", NULL, NULL, NULL, NULL);
+		break;
+	default:
+		break;
+	}
 
 	ofono_sim_set_data(sim, sd);
 	g_idle_add(at_sim_register, sim);
@@ -819,6 +818,9 @@ static void at_sim_remove(struct ofono_sim *sim)
 
 	ofono_sim_set_data(sim, NULL);
 
+	if (sd->epev_source)
+		g_source_remove(sd->epev_source);
+
 	g_free(sd);
 }
 
@@ -840,7 +842,6 @@ static struct ofono_sim_driver driver = {
 	.lock			= at_pin_enable,
 	.change_passwd		= at_change_passwd,
 	.query_locked		= at_pin_query_enabled,
-	.envelope		= at_sim_envelope,
 };
 
 void at_sim_init()
