@@ -2,7 +2,7 @@
  *
  *  oFono - Open Source Telephony
  *
- *  Copyright (C) 2008-2009  Intel Corporation. All rights reserved.
+ *  Copyright (C) 2008-2010  Intel Corporation. All rights reserved.
  *
  *  This program is free software; you can redistribute it and/or modify
  *  it under the terms of the GNU General Public License version 2 as
@@ -37,25 +37,24 @@
 #include <ofono/devinfo.h>
 #include <ofono/netreg.h>
 #include <ofono/sim.h>
-#include <ofono/sms.h>
+#include <ofono/stk.h>
 #include <ofono/cbs.h>
-#include <ofono/ssn.h>
+#include <ofono/sms.h>
 #include <ofono/ussd.h>
-#include <ofono/voicecall.h>
-#include <ofono/phonebook.h>
-#include <ofono/message-waiting.h>
-#include <ofono/call-meter.h>
-#include <ofono/call-settings.h>
-#include <ofono/call-volume.h>
 #include <ofono/gprs.h>
 #include <ofono/gprs-context.h>
 #include <ofono/log.h>
 
+#include <drivers/atmodem/vendor.h>
+
 static const char *cfun_prefix[] = { "+CFUN:", NULL };
+static const char *cpin_prefix[] = { "+CPIN:", NULL };
 static const char *none_prefix[] = { NULL };
 
 struct mbm_data {
-	GAtChat *chat;
+	GAtChat *modem_port;
+	GAtChat *data_port;
+	gboolean have_sim;
 };
 
 static int mbm_probe(struct ofono_modem *modem)
@@ -81,25 +80,45 @@ static void mbm_remove(struct ofono_modem *modem)
 
 	ofono_modem_set_data(modem, NULL);
 
-	g_at_chat_unref(data->chat);
+	g_at_chat_unref(data->data_port);
+	g_at_chat_unref(data->modem_port);
 	g_free(data);
 }
 
 static void mbm_debug(const char *str, void *user_data)
 {
-	ofono_info("%s", str);
+	const char *prefix = user_data;
+
+	ofono_info("%s %s", prefix, str);
+}
+
+static void simpin_check(gboolean ok, GAtResult *result, gpointer user_data)
+{
+	struct ofono_modem *modem = user_data;
+	struct mbm_data *data = ofono_modem_get_data(modem);
+
+	DBG("");
+
+	/* Modem returns error if there is no SIM in slot */
+	data->have_sim = ok;
+
+	ofono_modem_set_powered(modem, TRUE);
 }
 
 static void cfun_enable(gboolean ok, GAtResult *result, gpointer user_data)
 {
 	struct ofono_modem *modem = user_data;
+	struct mbm_data *data = ofono_modem_get_data(modem);
 
 	DBG("");
 
-	if (!ok)
+	if (!ok) {
 		ofono_modem_set_powered(modem, FALSE);
+		return;
+	}
 
-	ofono_modem_set_powered(modem, TRUE);
+	g_at_chat_send(data->modem_port, "AT+CPIN?", cpin_prefix,
+			simpin_check, modem, NULL);
 }
 
 static void cfun_query(gboolean ok, GAtResult *result, gpointer user_data)
@@ -122,12 +141,12 @@ static void cfun_query(gboolean ok, GAtResult *result, gpointer user_data)
 	g_at_result_iter_next_number(&iter, &status);
 
 	if (status == 4) {
-		g_at_chat_send(data->chat, "AT+CFUN=1", none_prefix,
+		g_at_chat_send(data->modem_port, "AT+CFUN=1", none_prefix,
 				cfun_enable, modem, NULL);
 		return;
 	}
 
-	ofono_modem_set_powered(modem, TRUE);
+	cfun_enable(TRUE, NULL, modem);
 }
 
 static void emrdy_notifier(GAtResult *result, gpointer user_data)
@@ -149,7 +168,7 @@ static void emrdy_notifier(GAtResult *result, gpointer user_data)
 	if (status != 1)
 		return;
 
-	g_at_chat_send(data->chat, "AT+CFUN?", cfun_prefix,
+	g_at_chat_send(data->modem_port, "AT+CFUN?", cfun_prefix,
 					cfun_query, modem, NULL);
 }
 
@@ -168,47 +187,73 @@ static void emrdy_query(gboolean ok, GAtResult *result, gpointer user_data)
 	 * EMRDY: 1 has been sent, in which case the emrdy_notifier should be
 	 * triggered eventually and we send CFUN? again.
 	 */
-	g_at_chat_send(data->chat, "AT+CFUN?", cfun_prefix,
+	g_at_chat_send(data->modem_port, "AT+CFUN?", cfun_prefix,
 					cfun_query, modem, NULL);
 };
 
+static GAtChat *create_port(const char *device)
+{
+	GAtSyntax *syntax;
+	GIOChannel *channel;
+	GAtChat *chat;
+
+	channel = g_at_tty_open(device, NULL);
+	if (!channel)
+		return NULL;
+
+	syntax = g_at_syntax_new_gsmv1();
+	chat = g_at_chat_new(channel, syntax);
+	g_at_syntax_unref(syntax);
+	g_io_channel_unref(channel);
+
+	if (!chat)
+		return NULL;
+
+	return chat;
+}
+
 static int mbm_enable(struct ofono_modem *modem)
 {
 	struct mbm_data *data = ofono_modem_get_data(modem);
-	GIOChannel *channel;
-	GAtSyntax *syntax;
-	const char *device;
+	const char *modem_dev;
+	const char *data_dev;
 
 	DBG("%p", modem);
 
-	device  = ofono_modem_get_string(modem, "ModemDevice");
-	if (!device) {
-		device = ofono_modem_get_string(modem, "Device");
-		if (!device)
-			return -EINVAL;
-	}
+	modem_dev = ofono_modem_get_string(modem, "ModemDevice");
+	data_dev = ofono_modem_get_string(modem, "DataDevice");
 
-	channel = g_at_tty_open(device, NULL);
-	if (!channel)
-		return -EIO;
+	DBG("%s, %s", modem_dev, data_dev);
 
-	syntax = g_at_syntax_new_gsmv1();
-	data->chat = g_at_chat_new(channel, syntax);
-	g_at_syntax_unref(syntax);
-	g_io_channel_unref(channel);
+	if (modem_dev == NULL || data_dev == NULL)
+		return -EINVAL;
 
-	if (!data->chat)
+	data->modem_port = create_port(modem_dev);
+
+	if (data->modem_port == NULL)
 		return -EIO;
 
 	if (getenv("OFONO_AT_DEBUG"))
-		g_at_chat_set_debug(data->chat, mbm_debug, NULL);
+		g_at_chat_set_debug(data->modem_port, mbm_debug, "Modem:");
 
-	g_at_chat_register(data->chat, "*EMRDY:", emrdy_notifier,
+	data->data_port = create_port(data_dev);
+
+	if (data->data_port == NULL) {
+		g_at_chat_unref(data->modem_port);
+		data->modem_port = NULL;
+
+		return -EIO;
+	}
+
+	if (getenv("OFONO_AT_DEBUG"))
+		g_at_chat_set_debug(data->data_port, mbm_debug, "Data:");
+
+	g_at_chat_register(data->modem_port, "*EMRDY:", emrdy_notifier,
 					FALSE, modem, NULL);
 
-	g_at_chat_send(data->chat, "AT&F E0 V1 X4 &C1 +CMEE=1", NULL,
+	g_at_chat_send(data->modem_port, "AT&F E0 V1 X4 &C1 +CMEE=1", NULL,
 					NULL, NULL, NULL);
-	g_at_chat_send(data->chat, "AT*EMRDY?", none_prefix,
+	g_at_chat_send(data->modem_port, "AT*EMRDY?", none_prefix,
 				emrdy_query, modem, NULL);
 
 	return -EINPROGRESS;
@@ -221,9 +266,11 @@ static void cfun_disable(gboolean ok, GAtResult *result, gpointer user_data)
 
 	DBG("");
 
-	g_at_chat_shutdown(data->chat);
-	g_at_chat_unref(data->chat);
-	data->chat = NULL;
+	g_at_chat_unref(data->modem_port);
+	data->modem_port = NULL;
+
+	g_at_chat_unref(data->data_port);
+	data->data_port = NULL;
 
 	if (ok)
 		ofono_modem_set_powered(modem, FALSE);
@@ -235,12 +282,12 @@ static int mbm_disable(struct ofono_modem *modem)
 
 	DBG("%p", modem);
 
-	if (!data->chat)
+	if (!data->modem_port)
 		return 0;
 
-	g_at_chat_cancel_all(data->chat);
-	g_at_chat_unregister_all(data->chat);
-	g_at_chat_send(data->chat, "AT+CFUN=4", NULL,
+	g_at_chat_cancel_all(data->modem_port);
+	g_at_chat_unregister_all(data->modem_port);
+	g_at_chat_send(data->modem_port, "AT+CFUN=4", NULL,
 					cfun_disable, modem, NULL);
 
 	return -EINPROGRESS;
@@ -249,43 +296,40 @@ static int mbm_disable(struct ofono_modem *modem)
 static void mbm_pre_sim(struct ofono_modem *modem)
 {
 	struct mbm_data *data = ofono_modem_get_data(modem);
+	struct ofono_sim *sim;
 
 	DBG("%p", modem);
 
-	ofono_devinfo_create(modem, 0, "atmodem", data->chat);
-	ofono_sim_create(modem, 0, "atmodem", data->chat);
-	ofono_voicecall_create(modem, 0, "atmodem", data->chat);
+	ofono_devinfo_create(modem, 0, "atmodem", data->modem_port);
+	sim = ofono_sim_create(modem, OFONO_VENDOR_MBM, "atmodem",
+				data->modem_port);
+
+	if (data->have_sim && sim)
+		ofono_sim_inserted_notify(sim, TRUE);
 }
 
 static void mbm_post_sim(struct ofono_modem *modem)
 {
 	struct mbm_data *data = ofono_modem_get_data(modem);
-	struct ofono_message_waiting *mw;
 	struct ofono_gprs *gprs;
 	struct ofono_gprs_context *gc;
 
 	DBG("%p", modem);
 
-	ofono_call_settings_create(modem, 0, "atmodem", data->chat);
-	ofono_call_meter_create(modem, 0, "atmodem", data->chat);
-	ofono_call_volume_create(modem, 0, "atmodem", data->chat);
+	ofono_stk_create(modem, 0, "mbmmodem", data->modem_port);
 
-	ofono_ussd_create(modem, 0, "atmodem", data->chat);
-	ofono_netreg_create(modem, 0, "atmodem", data->chat);
-	ofono_phonebook_create(modem, 0, "atmodem", data->chat);
-	ofono_ssn_create(modem, 0, "atmodem", data->chat);
-	ofono_sms_create(modem, 0, "atmodem", data->chat);
-	ofono_cbs_create(modem, 0, "atmodem", data->chat);
+	ofono_netreg_create(modem, OFONO_VENDOR_MBM, "atmodem",
+				data->modem_port);
 
-	gprs = ofono_gprs_create(modem, 0, "atmodem", data->chat);
-	gc = ofono_gprs_context_create(modem, 0, "mbm", data->chat);
+	ofono_sms_create(modem, 0, "atmodem", data->modem_port);
+	ofono_cbs_create(modem, 0, "atmodem", data->modem_port);
+	ofono_ussd_create(modem, 0, "atmodem", data->modem_port);
+
+	gprs = ofono_gprs_create(modem, 0, "atmodem", data->modem_port);
+	gc = ofono_gprs_context_create(modem, 0, "mbm", data->modem_port);
 
 	if (gprs && gc)
 		ofono_gprs_add_context(gprs, gc);
-
-	mw = ofono_message_waiting_create(modem);
-	if (mw)
-		ofono_message_waiting_register(mw);
 }
 
 static struct ofono_modem_driver mbm_driver = {
