@@ -64,9 +64,12 @@ struct ofono_sms {
 	struct ofono_sim *sim;
 	GKeyFile *settings;
 	char *imsi;
+	int bearer;
 	const struct ofono_sms_driver *driver;
 	void *driver_data;
 	struct ofono_atom *atom;
+	ofono_bool_t use_delivery_reports;
+	struct status_report_assembly *sr_assembly;
 };
 
 struct pending_pdu {
@@ -82,8 +85,59 @@ struct tx_queue_entry {
 	unsigned int msg_id;
 	unsigned int retry;
 	DBusMessage *msg;
+	gboolean status_report;
+	struct sms_address receiver;
 };
 
+static const char *sms_bearer_to_string(int bearer)
+{
+	switch (bearer) {
+	case 0:
+		return "ps-only";
+	case 1:
+		return "cs-only";
+	case 2:
+		return "ps-preferred";
+	case 3:
+		return "cs-preferred";
+	};
+
+	return "unknown";
+}
+
+static int sms_bearer_from_string(const char *str)
+{
+	if (g_str_equal(str, "ps-only"))
+		return 0;
+	else if (g_str_equal(str, "cs-only"))
+		return 1;
+	else if (g_str_equal(str, "ps-preferred"))
+		return 2;
+	else if (g_str_equal(str, "cs-preferred"))
+		return 3;
+
+	return -1;
+}
+
+static void set_bearer(struct ofono_sms *sms, int bearer)
+{
+	DBusConnection *conn = ofono_dbus_get_connection();
+	const char *path = __ofono_atom_get_path(sms->atom);
+	const char *value;
+
+	if (sms->bearer == bearer)
+		return;
+
+	sms->bearer = bearer;
+
+	value = sms_bearer_to_string(sms->bearer);
+
+	ofono_dbus_signal_property_changed(conn, path,
+						OFONO_SMS_MANAGER_INTERFACE,
+						"Bearer",
+						DBUS_TYPE_STRING, &value);
+}
+
 static void set_sca(struct ofono_sms *sms,
 			const struct ofono_phone_number *sca)
 {
@@ -114,6 +168,7 @@ static DBusMessage *generate_get_properties_reply(struct ofono_sms *sms,
 	DBusMessageIter iter;
 	DBusMessageIter dict;
 	const char *sca;
+	const char *bearer;
 
 	reply = dbus_message_new_method_return(msg);
 
@@ -131,13 +186,20 @@ static DBusMessage *generate_get_properties_reply(struct ofono_sms *sms,
 	ofono_dbus_dict_append(&dict, "ServiceCenterAddress", DBUS_TYPE_STRING,
 				&sca);
 
+	ofono_dbus_dict_append(&dict, "UseDeliveryReports", DBUS_TYPE_BOOLEAN,
+				&sms->use_delivery_reports);
+
+	bearer = sms_bearer_to_string(sms->bearer);
+	ofono_dbus_dict_append(&dict, "Bearer", DBUS_TYPE_STRING, &bearer);
+
 	dbus_message_iter_close_container(&iter, &dict);
 
 	return reply;
 }
 
 static void sms_sca_query_cb(const struct ofono_error *error,
-				const struct ofono_phone_number *sca, void *data)
+				const struct ofono_phone_number *sca,
+				void *data)
 {
 	struct ofono_sms *sms = data;
 
@@ -177,6 +239,39 @@ static DBusMessage *sms_get_properties(DBusConnection *conn,
 	return NULL;
 }
 
+static void bearer_set_query_callback(const struct ofono_error *error,
+					int bearer, void *data)
+{
+	struct ofono_sms *sms = data;
+	DBusMessage *reply;
+
+	if (error->type != OFONO_ERROR_TYPE_NO_ERROR) {
+		ofono_error("Set Bearer succeeded, but query failed");
+		reply = __ofono_error_failed(sms->pending);
+		__ofono_dbus_pending_reply(&sms->pending, reply);
+		return;
+	}
+
+	reply = dbus_message_new_method_return(sms->pending);
+	__ofono_dbus_pending_reply(&sms->pending, reply);
+
+	set_bearer(sms, bearer);
+}
+
+static void bearer_set_callback(const struct ofono_error *error, void *data)
+{
+	struct ofono_sms *sms = data;
+
+	if (error->type != OFONO_ERROR_TYPE_NO_ERROR) {
+		DBG("Setting Bearer failed");
+		__ofono_dbus_pending_reply(&sms->pending,
+					__ofono_error_failed(sms->pending));
+		return;
+	}
+
+	sms->driver->bearer_query(sms, bearer_set_query_callback, sms);
+}
+
 static void sca_set_query_callback(const struct ofono_error *error,
 					const struct ofono_phone_number *sca,
 					void *data)
@@ -249,7 +344,8 @@ static DBusMessage *sms_set_property(DBusConnection *conn, DBusMessage *msg,
 		if (strlen(value) == 0 || !valid_phone_number_format(value))
 			return __ofono_error_invalid_format(msg);
 
-		if (!sms->driver->sca_set)
+		if (sms->driver->sca_set == NULL ||
+				sms->driver->sca_query == NULL)
 			return __ofono_error_not_implemented(msg);
 
 		string_to_phone_number(value, &sca);
@@ -260,6 +356,51 @@ static DBusMessage *sms_set_property(DBusConnection *conn, DBusMessage *msg,
 		return NULL;
 	}
 
+	if (!strcmp(property, "Bearer")) {
+		const char *value;
+		int bearer;
+
+		if (dbus_message_iter_get_arg_type(&var) != DBUS_TYPE_STRING)
+			return __ofono_error_invalid_args(msg);
+
+		dbus_message_iter_get_basic(&var, &value);
+
+		bearer = sms_bearer_from_string(value);
+		if (bearer < 0)
+			return __ofono_error_invalid_format(msg);
+
+		if (sms->driver->bearer_set == NULL ||
+				sms->driver->bearer_query == NULL)
+			return __ofono_error_not_implemented(msg);
+
+		sms->pending = dbus_message_ref(msg);
+
+		sms->driver->bearer_set(sms, bearer, bearer_set_callback, sms);
+		return NULL;
+	}
+
+	if (!strcmp(property, "UseDeliveryReports")) {
+		const char *path = __ofono_atom_get_path(sms->atom);
+		dbus_bool_t value;
+
+		if (dbus_message_iter_get_arg_type(&var) != DBUS_TYPE_BOOLEAN)
+			return __ofono_error_invalid_args(msg);
+
+		dbus_message_iter_get_basic(&var, &value);
+
+		g_dbus_send_reply(conn, msg, DBUS_TYPE_INVALID);
+
+		if (sms->use_delivery_reports != (ofono_bool_t) value) {
+			sms->use_delivery_reports = value;
+			ofono_dbus_signal_property_changed(conn, path,
+						OFONO_SMS_MANAGER_INTERFACE,
+						"UseDeliveryReports",
+						DBUS_TYPE_BOOLEAN, &value);
+		}
+
+		return NULL;
+	}
+
 	return __ofono_error_invalid_args(msg);
 }
 
@@ -306,6 +447,13 @@ static void tx_finished(const struct ofono_error *error, int mr, void *data)
 	entry->cur_pdu += 1;
 	entry->retry = 0;
 
+	if (entry->status_report)
+		status_report_assembly_add_fragment(sms->sr_assembly,
+							entry->msg_id,
+							&entry->receiver,
+							mr, time(NULL),
+							entry->num_pdus);
+
 	if (entry->cur_pdu < entry->num_pdus) {
 		sms->tx_source = g_timeout_add(0, tx_next, sms);
 		return;
@@ -426,7 +574,8 @@ static DBusMessage *sms_send_message(DBusConnection *conn, DBusMessage *msg,
 	if (valid_phone_number_format(to) == FALSE)
 		return __ofono_error_invalid_format(msg);
 
-	msg_list = sms_text_prepare(text, 0, TRUE, &ref_offset);
+	msg_list = sms_text_prepare(text, 0, TRUE, &ref_offset,
+					sms->use_delivery_reports);
 
 	if (!msg_list)
 		return __ofono_error_invalid_format(msg);
@@ -436,6 +585,8 @@ static DBusMessage *sms_send_message(DBusConnection *conn, DBusMessage *msg,
 	set_ref_and_to(msg_list, sms->ref, ref_offset, to);
 	entry = create_tx_queue_entry(msg_list);
 
+	sms_address_from_string(&entry->receiver, to);
+
 	g_slist_foreach(msg_list, (GFunc)g_free, NULL);
 	g_slist_free(msg_list);
 
@@ -448,6 +599,7 @@ static DBusMessage *sms_send_message(DBusConnection *conn, DBusMessage *msg,
 
 	entry->msg = dbus_message_ref(msg);
 	entry->msg_id = sms->next_msg_id++;
+	entry->status_report = sms->use_delivery_reports;
 
 	g_queue_push_tail(sms->txq, entry);
 
@@ -692,6 +844,23 @@ static void handle_deliver(struct ofono_sms *sms, const struct sms *incoming)
 	g_slist_free(l);
 }
 
+static void handle_sms_status_report(struct ofono_sms *sms,
+						const struct sms *incoming)
+{
+	struct ofono_modem *modem = __ofono_atom_get_modem(sms->atom);
+	gboolean delivered;
+	unsigned int msg_id;
+
+	if (status_report_assembly_report(sms->sr_assembly, incoming, &msg_id,
+						&delivered) == FALSE)
+		return;
+
+	__ofono_history_sms_send_status(modem, msg_id, time(NULL),
+			delivered ? OFONO_HISTORY_SMS_STATUS_DELIVERED :
+			OFONO_HISTORY_SMS_STATUS_DELIVER_FAILED);
+}
+
+
 static inline gboolean handle_mwi(struct ofono_sms *sms, struct sms *s)
 {
 	gboolean discard;
@@ -823,7 +992,30 @@ out:
 void ofono_sms_status_notify(struct ofono_sms *sms, unsigned char *pdu,
 				int len, int tpdu_len)
 {
-	ofono_error("SMS Status-Report not yet handled");
+	struct sms s;
+	enum sms_class cls;
+
+	if (!sms_decode(pdu, len, FALSE, tpdu_len, &s)) {
+		ofono_error("Unable to decode PDU");
+		return;
+	}
+
+	if (s.type != SMS_TYPE_STATUS_REPORT) {
+		ofono_error("Expecting a STATUS REPORT pdu");
+		return;
+	}
+
+	if (s.status_report.srq) {
+		ofono_error("Waiting an answer to SMS-SUBMIT, not SMS-COMMAND");
+		return;
+	}
+
+	if (!sms_dcs_decode(s.deliver.dcs, &cls, NULL, NULL, NULL)) {
+		ofono_error("Unknown / Reserved DCS.  Ignoring");
+		return;
+	}
+
+	handle_sms_status_report(sms, &s);
 }
 
 int ofono_sms_driver_register(const struct ofono_sms_driver *d)
@@ -895,6 +1087,11 @@ static void sms_remove(struct ofono_atom *atom)
 					"NextMessageId", sms->next_msg_id);
 		g_key_file_set_integer(sms->settings, SETTINGS_GROUP,
 					"NextReference", sms->ref);
+		g_key_file_set_boolean(sms->settings, SETTINGS_GROUP,
+					"UseDeliveryReports",
+					sms->use_delivery_reports);
+		g_key_file_set_integer(sms->settings, SETTINGS_GROUP,
+					"Bearer", sms->bearer);
 
 		storage_close(sms->imsi, SETTINGS_STORE, sms->settings, TRUE);
 
@@ -903,6 +1100,11 @@ static void sms_remove(struct ofono_atom *atom)
 		sms->settings = NULL;
 	}
 
+	if (sms->sr_assembly) {
+		status_report_assembly_free(sms->sr_assembly);
+		sms->sr_assembly = NULL;
+	}
+
 	g_free(sms);
 }
 
@@ -970,6 +1172,8 @@ static void mw_watch(struct ofono_atom *atom,
 
 static void sms_load_settings(struct ofono_sms *sms, const char *imsi)
 {
+	GError *error = NULL;
+
 	sms->settings = storage_open(imsi, SETTINGS_STORE);
 
 	if (sms->settings == NULL)
@@ -979,14 +1183,27 @@ static void sms_load_settings(struct ofono_sms *sms, const char *imsi)
 
 	sms->next_msg_id = g_key_file_get_integer(sms->settings, SETTINGS_GROUP,
 							"NextMessageId", NULL);
+
 	sms->ref = g_key_file_get_integer(sms->settings, SETTINGS_GROUP,
 							"NextReference", NULL);
-
 	if (sms->ref >= 65536)
 		sms->ref = 1;
 
+	sms->use_delivery_reports =
+		g_key_file_get_boolean(sms->settings, SETTINGS_GROUP,
+					"UseDeliveryReports", NULL);
+
+	sms->bearer = g_key_file_get_integer(sms->settings, SETTINGS_GROUP,
+							"Bearer", &error);
+	if (error)
+		sms->bearer = 3; /* Default to CS then PS */
 }
 
+static void bearer_init_callback(const struct ofono_error *error, void *data)
+{
+	if (error->type != OFONO_ERROR_TYPE_NO_ERROR)
+		ofono_error("Error bootstrapping SMS Bearer Preference");
+}
 
 /*
  * Indicate oFono that a SMS driver is ready for operation
@@ -1038,11 +1255,19 @@ void ofono_sms_register(struct ofono_sms *sms)
 		imsi = ofono_sim_get_imsi(sms->sim);
 		sms->assembly = sms_assembly_new(imsi);
 
+		sms->sr_assembly = status_report_assembly_new(imsi);
+
 		sms_load_settings(sms, imsi);
 	} else {
 		sms->assembly = sms_assembly_new(NULL);
+		sms->sr_assembly = status_report_assembly_new(NULL);
+		sms->bearer = 3; /* Default to CS then PS */
 	}
 
+	if (sms->driver->bearer_set)
+		sms->driver->bearer_set(sms, sms->bearer,
+						bearer_init_callback, sms);
+
 	__ofono_atom_register(sms->atom, sms_unregister);
 }
 
