@@ -26,6 +26,7 @@
 #define _GNU_SOURCE
 #include <string.h>
 #include <stdio.h>
+#include <stdint.h>
 
 #include <glib.h>
 #include <gdbus.h>
@@ -85,24 +86,28 @@ void ofono_stk_proactive_command_notify(struct ofono_stk *stk,
 	char *buf;
 	int i;
 
-	buf = g_try_new(char, length * 2 + 1);
+	buf = g_try_malloc(length * 2 + 1);
+	if (!buf)
+		return;
+
 	for (i = 0; i < length; i ++)
 		sprintf(buf + i * 2, "%02hhx", pdu[i]);
-	DBG("Proactive command PDU: %s", buf);
 
 	cmd = stk_command_new_from_pdu(pdu, length);
 	if (!cmd) {
 		ofono_error("Can't parse proactive command: %s", buf);
-		g_free(buf);
 
 		/* TODO: return TERMINAL RESPONSE with permanent error */
-		return;
+		goto done;
 	}
 
 	/* TODO: execute */
+	ofono_info("Proactive command PDU: %s", buf);
 
-	g_free(buf);
 	stk_command_free(cmd);
+
+done:
+	g_free(buf);
 }
 
 int ofono_stk_driver_register(const struct ofono_stk_driver *d)
